import os
import fnmatch
from typing import List


def read_crllm_ignore(project_root: str) -> List[str]:
    """
    Read .crllm_ignore file and return a list of ignore patterns.

    Args:
        project_root (str): Root directory of the project

    Returns:
        List[str]: List of ignore patterns
    """
    ignore_file_path = os.path.join(project_root, ".crllm_ignore")

    if not os.path.exists(ignore_file_path):
        return []

    with open(ignore_file_path, "r", encoding="utf8") as file:
        # Strip whitespace and remove comments and empty lines
        return [
            line.strip()
            for line in file.readlines()
            if line.strip() and not line.strip().startswith("#")
        ]


def should_ignore_file(
    file_path: str, project_root: str, ignore_patterns: List[str] = None
) -> bool:

    if ignore_patterns is None:
        ignore_patterns = read_crllm_ignore(project_root)

    relative_path = os.path.relpath(file_path, project_root).replace(os.path.sep, "/")

    return any(
        fnmatch.fnmatch(relative_path, pattern)
        or fnmatch.fnmatch(os.path.basename(file_path), pattern)
        for pattern in ignore_patterns
    )



import argparse
import logging
import os
from crllm.init_project import init_project
from crllm.app import app
from crllm.config.config_service import config_service


def cli():
    log_level = os.environ.get("LOGLEVEL", "INFO").upper()
    logging.basicConfig(level=log_level, format="%(asctime)s %(message)s")

    parser = argparse.ArgumentParser(
        description="Get Code Reviews from large language models."
    )

    parser.add_argument("input", help="Path to the sources")
    parser.add_argument(
        "-c", "--config", help="Path to the config file", default="crllm_config.toml"
    )

    parser.add_argument(
        "-l",
        "--loader",
        help="Loader to use",
        required=False,
        choices=["file", "git", "git_compare"],
    )

    parser.add_argument(
        "-i", "--init", help="Initialize the config file", action="store_true"
    )

    parser.add_argument(
        "-o",
        "--output",
        help="Output file for the generated code review",
        required=False,
        default="",
    )

    args = parser.parse_args()

    config = {"crllm": {"output": args.output}}

    if args.loader:
        config["crllm"]["loader"] = args.loader

    config_service.set_config_path(args.config)
    config_service.override_config(config)

    if args.init:
        init_project(args.input)
        return

    app(args.input)


if __name__ == "__main__":
    cli()



from langchain_openai import ChatOpenAI, OpenAIEmbeddings
from crllm.model.model import Model


class OpenAIModel(Model):
    def _get_model(self, model_config):
        return ChatOpenAI(**model_config)

    def _get_embeddings(self, embeddings_config):
        return OpenAIEmbeddings(**embeddings_config)

    @staticmethod
    def get_required_config():
        return ["model"]



import pytest
from crllm.config.config_service import ConfigService


@pytest.fixture()
def mock_config_file(tmpdir):
    tmpdir.join("config.toml").write(
        """
    [crllm]
        loader = "foo"
        provider = "bar"
    """
    )

    return str(tmpdir.join("config.toml"))


def test_get_config(mock_config_file):
    config_service = ConfigService()
    config_service.set_config_path(mock_config_file)

    result = config_service.get_config()

    assert result["crllm"]["loader"] == "foo"
    assert result["crllm"]["provider"] == "bar"


def test_override(mock_config_file):
    config_service = ConfigService()
    config_service.set_config_path(mock_config_file)

    config_service.override_config({"crllm": {"loader": "test"}})

    result = config_service.get_config()

    assert result["crllm"]["loader"] == "test"



from crllm.code import code_loader
from crllm.code.loaders.file import file
from crllm.code.loaders.git_compare import git_compare
from crllm.code.loaders.git import git

loaders = {"git": git, "git_compare": git_compare, "file": file}



from git import Repo
from crllm.code.loaders.ignore_utils import should_ignore_file
from crllm.config.config_service import config_service


def git_compare(path):
    repo = Repo(path)

    current_branch = repo.active_branch.name
    config = config_service.get_config()
    main_branch = config["crllm"]["git_main_branch"]

    diff = repo.git.diff(f"{main_branch}...{current_branch}")

    code = ""

    for line in diff.splitlines():
        if "b/" in line:
            file = line.split("b/")[1]
            full_file_path = f"{path}/{file}"

            if should_ignore_file(full_file_path, path):
                continue

            code += "\nFile name: " + file + "\n"

        code += line + "\n"

    return code



from crllm.config.config_service import config_service
from crllm.model.model import Model
from crllm.model.provider import providers


class ModelFactory:
    def get_model(self) -> Model:
        config = config_service.get_config()
        provider = config["crllm"]["provider"]

        return providers[provider]()



from crllm.model.model import Model


class DryRunModel(Model):
    def _get_model(self, model_config):
        pass

    def _get_embeddings(self, embeddings_config):
        pass

    def generate(self, prompt_template, prompt_args):
        return "This is a dry run!"



import logging
from crllm.code import loaders
from crllm.config.config_service import config_service


class CodeLoader:
    def __init__(self) -> None:
        self.loaders = {}

    def get_code(self, path):
        config = config_service.get_config()

        loader_name = config["crllm"]["loader"]

        result = loaders.loaders[loader_name](path)
        logging.debug(result)
        return result


loader = CodeLoader()



import os
import tempfile
from crllm.code.loaders.ignore_utils import read_crllm_ignore, should_ignore_file


def test_empty_file():
    with tempfile.TemporaryDirectory() as tmpdir:
        ignore_path = os.path.join(tmpdir, ".crllm_ignore")
        with open(ignore_path, "w", encoding="utf8") as file:
            file.close()
        assert read_crllm_ignore(tmpdir) == []


def test_comments_and_whitespace():
    with tempfile.TemporaryDirectory() as tmpdir:
        ignore_path = os.path.join(tmpdir, ".crllm_ignore")
        with open(ignore_path, "w", encoding="utf8") as file:
            file.write(
                """
# This is a comment
*.log
 temp/  
# Another comment
.env
"""
            )
        patterns = read_crllm_ignore(tmpdir)
        assert patterns == ["*.log", "temp/", ".env"]


def test_read_crllm_ignore_no_file():
    with tempfile.TemporaryDirectory() as tmpdir:
        assert read_crllm_ignore(tmpdir) == []


def test_ignore_relative_path():
    with tempfile.TemporaryDirectory() as tmpdir:

        os.makedirs(os.path.join(tmpdir, "src", "logs"))
        ignore_path = os.path.join(tmpdir, ".crllm_ignore")
        with open(ignore_path, "w", encoding="utf8") as file:
            file.write("*.log\nsrc/logs/\n*.pyc\ntemp/")

        assert should_ignore_file(os.path.join(tmpdir, "app.log"), tmpdir)
        assert should_ignore_file(
            os.path.join(tmpdir, "src", "logs", "debug.log"), tmpdir
        )
        assert not should_ignore_file(os.path.join(tmpdir, "src", "main.py"), tmpdir)


def test_basename_patterns():
    with tempfile.TemporaryDirectory() as tmpdir:
        ignore_path = os.path.join(tmpdir, ".crllm_ignore")
        with open(ignore_path, "w", encoding="utf8") as file:
            file.write(".env\n*.log")

        assert should_ignore_file(os.path.join(tmpdir, "src", ".env"), tmpdir)
        assert should_ignore_file(os.path.join(tmpdir, "app.log"), tmpdir)
        assert should_ignore_file(os.path.join(tmpdir, "src", "app.log"), tmpdir)


def test_custom_patterns():
    with tempfile.TemporaryDirectory() as tmpdir:
        os.makedirs(os.path.join(tmpdir, "src"))

        custom_patterns = ["*.txt", "secret/"]
        file_path = os.path.join(tmpdir, "src", "notes.txt")
        assert should_ignore_file(file_path, tmpdir, custom_patterns)

        file_path = os.path.join(tmpdir, "src", "code.py")
        assert not should_ignore_file(file_path, tmpdir, custom_patterns)


def test_edge_cases():
    with tempfile.TemporaryDirectory() as tmpdir:
        assert not should_ignore_file(os.path.join(tmpdir, "test.py"), tmpdir, [])

        ignore_path = os.path.join(tmpdir, ".crllm_ignore")
        with open(ignore_path, "w", encoding="utf8") as file:
            file.write("*.tmp")

        assert should_ignore_file(os.path.join(tmpdir, "test.tmp"), tmpdir)



from langchain_huggingface import (
    ChatHuggingFace,
    HuggingFaceEndpoint,
    HuggingFaceEmbeddings,
)
from crllm.model.model import Model


class HuggingFaceModel(Model):
    def _get_model(self, model_config):
        llm = HuggingFaceEndpoint(**model_config)

        return ChatHuggingFace(llm=llm)

    def _get_embeddings(self, embeddings_config):
        return HuggingFaceEmbeddings(**embeddings_config)

    @staticmethod
    def get_required_config():
        return ["repo_id", "task"]



import logging
from langchain_core.prompts import ChatPromptTemplate
from crllm.config.config_service import config_service


class PromptBuilder:
    def build(self):
        config = config_service.get_config()

        logging.debug(config["prompt"])

        messages = [
            ("system", config["prompt"]["template"]),
            ("human", "Do code review for the following code changes: \n {code}"),
        ]

        if config["rag"]["enabled"]:
            self._insert_rag_message(messages)

        return ChatPromptTemplate.from_messages(messages)

    def _insert_rag_message(self, messages):
        messages.insert(
            1,
            (
                "system",
                """
This code could be relevant to understand the changes:
<context>
    {context}
</context>
                    """,
            ),
        )



import logging
from rich.console import Console
from rich.markdown import Markdown
from crllm.config.config_service import config_service
from crllm.code import code_loader
from crllm.model.model_factory import ModelFactory
from crllm.prompt_builder import PromptBuilder


def app(path: str):
    logging.info("Reading %s", path)

    code = code_loader.loader.get_code(path)

    prompt_builder = PromptBuilder()
    prompt = prompt_builder.build()

    model_factory = ModelFactory()

    model = model_factory.get_model()

    result = model.generate(prompt, {"code": code})

    markdown = Markdown(result)

    config = config_service.get_config()

    if config["crllm"]["output"] != "":
        with open(config["crllm"]["output"], "w", encoding="utf8") as file:
            file.write(str(result))

    console = Console()
    console.print(markdown)



import sys
from crllm.cli import cli
from crllm.code.loaders.file import file


def test_cli(monkeypatch, capfd, tmpdir):

    file_path = "./src/crllm/test/functional/smoke_test.py"
    config_path = "./src/crllm/test/functional/config.toml"
    output_path = str(tmpdir.join("output.md"))

    monkeypatch.setattr(
        sys, "argv", ["crllm", file_path, "-c", config_path, "-o", output_path]
    )
    cli()
    out, err = capfd.readouterr()

    assert not err
    assert "This is a dry run!" in out
    assert file(output_path) == "This is a dry run!"
